import unittest
import search

from assignment import AssignmentProblem, AssignmentSolution

class TestAssignment(unittest.TestCase):

    def test_problem(self):
        problem = AssignmentProblem([[94,1,54,68],[74,10,88,82],[62,88,8,76],[11,74,81,21]])
        self.assertEqual(4, problem.size())

    def test_root(self):
        problem = AssignmentProblem([[94,1,54,68],[74,10,88,82],[62,88,8,76],[11,74,81,21]])
        root = AssignmentSolution(problem)
        self.assertEqual(0, root.size())
        self.assertFalse(root.solution())
        self.assertEqual(0, root.eval())
        self.assertEqual(41, root.lower_bound())
        self.assertEqual(102, root.upper_bound())

    def test_root_children(self):
        problem = AssignmentProblem([[94,1,54,68],[74,10,88,82],[62,88,8,76],[11,74,81,21]])
        root = AssignmentSolution(problem)
        children = root.children()
        self.assertEqual(4, len(children))

        self.assertEqual(1, children[0].size())
        self.assertEqual(0, children[0].assignment[0])
        self.assertEqual(94, children[0].eval())

        self.assertEqual(1, children[1].size())
        self.assertEqual(1, children[1].assignment[0])
        self.assertEqual(74, children[1].eval())
        
        self.assertEqual(1, children[2].size())
        self.assertEqual(2, children[2].assignment[0])
        self.assertEqual(62, children[2].eval())
        
        self.assertEqual(1, children[3].size())
        self.assertEqual(3, children[3].assignment[0])
        self.assertEqual(11, children[3].eval())

    def test_solution_greedy(self):
        problem = AssignmentProblem([[94,1,54,68],[74,10,88,82],[62,88,8,76],[11,74,81,21]])
        root = AssignmentSolution(problem)
        greedy = root.greedy()
        print(greedy)
        self.assertEqual(4, greedy.size())
        self.assertTrue(greedy.solution())
        self.assertEqual(102, greedy.eval())
        self.assertEqual(102, greedy.lower_bound())
        self.assertEqual(102, greedy.upper_bound())
        self.assertEqual(3, greedy.assignment[0])
        self.assertEqual(0, greedy.assignment[1])
        self.assertEqual(2, greedy.assignment[2])
        self.assertEqual(1, greedy.assignment[3])

    def test_solution_branch_and_bound(self):
        problem = AssignmentProblem([[94,1,54,68],[74,10,88,82],[62,88,8,76],[11,74,81,21]])
        root = AssignmentSolution(problem)
        (c,s) = search.branch_and_bound_min(root)
        self.assertEqual(97, c)
        self.assertEqual(4, s.size())
        self.assertTrue(s.solution())
        self.assertEqual(97, s.eval())
        self.assertEqual(97, s.lower_bound())
        self.assertEqual(97, s.upper_bound())
        self.assertEqual(3, s.assignment[0])
        self.assertEqual(1, s.assignment[1])
        self.assertEqual(2, s.assignment[2])
        self.assertEqual(0, s.assignment[3])

    # Maximize benefits == Minimize costs (where cost = -benefit)

    def test_problem_max(self):
        problem = AssignmentProblem([[-5,-6,-4],[-3,-8,-2],[-6,-5,-1]])
        self.assertEqual(3, problem.size())

    def test_root_max(self):
        problem = AssignmentProblem([[-5,-6,-4],[-3,-8,-2],[-6,-5,-1]])
        root = AssignmentSolution(problem)
        self.assertEqual(0, root.size())
        self.assertFalse(root.solution())
        self.assertEqual(0, root.eval())
        self.assertEqual(-18, root.lower_bound())
        self.assertEqual(-18, root.upper_bound())

    def test_root_children_max(self):
        problem = AssignmentProblem([[-5,-6,-4],[-3,-8,-2],[-6,-5,-1]])
        root = AssignmentSolution(problem)
        children = root.children()
        self.assertEqual(3, len(children))

        self.assertEqual(1, children[0].size())
        self.assertEqual(0, children[0].assignment[0])
        self.assertEqual(-5, children[0].eval())

        self.assertEqual(1, children[1].size())
        self.assertEqual(1, children[1].assignment[0])
        self.assertEqual(-3, children[1].eval())
        
        self.assertEqual(1, children[2].size())
        self.assertEqual(2, children[2].assignment[0])
        self.assertEqual(-6, children[2].eval())

    def test_solution_greedy_max(self):
        problem = AssignmentProblem([[-5,-6,-4],[-3,-8,-2],[-6,-5,-1]])
        root = AssignmentSolution(problem)
        greedy = root.greedy()
        print(greedy)
        self.assertEqual(3, greedy.size())
        self.assertTrue(greedy.solution())
        self.assertEqual(-18, greedy.eval())
        self.assertEqual(-18, greedy.lower_bound())
        self.assertEqual(-18, greedy.upper_bound())
        self.assertEqual(2, greedy.assignment[0])
        self.assertEqual(1, greedy.assignment[1])
        self.assertEqual(0, greedy.assignment[2])

    def test_solution_branch_and_bound_max(self):
        problem = AssignmentProblem([[-5,-6,-4],[-3,-8,-2],[-6,-5,-1]])
        root = AssignmentSolution(problem)
        (c,s) = search.branch_and_bound_min(root)
        self.assertEqual(-18, c)
        self.assertEqual(3, s.size())
        self.assertTrue(s.solution())
        self.assertEqual(-18, s.eval())
        self.assertEqual(-18, s.lower_bound())
        self.assertEqual(-18, s.upper_bound())
        self.assertEqual(2, s.assignment[0])
        self.assertEqual(1, s.assignment[1])
        self.assertEqual(0, s.assignment[2])


if __name__ == '__main__':
    unittest.main()

 

print("ASSIGNMENT PROBLEM MAX (COSTS = -BENEFITS)")
problem = AssignmentProblem([[-5,-6,-4],[-3,-8,-2],[-6,-5,-1]])
print(problem)

print("Root")
print(root)
print("Greedy solution")
print(root.greedy())

(c,s) = branch_and_bound_min(root)
print(f"Best cost = {c}")
print(f"Best solution = {s}")

(c,s) = branch_and_bound_min_no_greedy(root)
print(f"Best cost = {c}")
print(f"Best solution = {s}")


