import unittest
import search

from knapsack import KnapsackProblem, KnapsackSolution

class TestKnapsack(unittest.TestCase):

    def setUp(self):
        self.problem = KnapsackProblem(61, [1,11,21,23,33], [11,21,31,33,43])

    def test_problem(self):
        self.assertEqual(5, self.problem.size())
        self.assertEqual(61, self.problem.capacity)
        self.assertEqual([1,11,21,23,33], self.problem.weights)
        self.assertEqual([11,21,31,33,43], self.problem.benefits)

    def test_root(self):
        root = KnapsackSolution(self.problem)
        print(root)
        self.assertEqual(0, root.size())
        self.assertFalse(root.solution())
        self.assertEqual(0, root.eval())
        self.assertAlmostEqual( 96.00, root.estimation(), delta=0.01)
        self.assertAlmostEqual(  0.00, root.lower_bound(), delta=0.01)
        self.assertAlmostEqual(102.52, root.upper_bound(), delta=0.01)

    def test_root_children(self):
        root = KnapsackSolution(self.problem)
        children = root.children()
        self.assertEqual(2, len(children))

        self.assertEqual(1, children[0].size())
        self.assertEqual(0, children[0].assignment[0])
        self.assertEqual(0, children[0].eval())
        self.assertEqual(85, children[0].estimation())

        self.assertEqual(1, children[1].size())
        self.assertEqual(1, children[1].assignment[0])
        self.assertEqual(11, children[1].eval())
        self.assertEqual(96, children[1].estimation())

    def test_solution_greedy(self):
        root = KnapsackSolution(self.problem)
        greedy_solution, greedy_value = root.greedy()
        self.assertAlmostEqual(102.52, greedy_value, delta=0.01)
        self.assertEqual(5, greedy_solution.size())
        self.assertTrue(greedy_solution.solution())
        self.assertEqual(56, greedy_solution.weight())
        self.assertEqual(96, greedy_solution.eval())
        self.assertEqual(96, greedy_solution.lower_bound())
        self.assertEqual(96, greedy_solution.upper_bound())
        self.assertEqual(1, greedy_solution.assignment[0])
        self.assertEqual(1, greedy_solution.assignment[1])
        self.assertEqual(1, greedy_solution.assignment[2])
        self.assertEqual(1, greedy_solution.assignment[3])
        self.assertEqual(0, greedy_solution.assignment[4])

    def test_solution_branch_and_bound(self):
        root = KnapsackSolution(self.problem)
        (benefit,solution) = search.branch_and_bound_max(root)
        self.assertEqual(96, benefit)
        self.assertEqual(5, solution.size())
        self.assertTrue(solution.solution())
        self.assertEqual(56, solution.weight())
        self.assertEqual(96, solution.eval())
        self.assertEqual(96, solution.lower_bound())
        self.assertEqual(96, solution.upper_bound())
        self.assertEqual(1, solution.assignment[0])
        self.assertEqual(1, solution.assignment[1])
        self.assertEqual(1, solution.assignment[2])
        self.assertEqual(1, solution.assignment[3])
        self.assertEqual(0, solution.assignment[4])

    # Problem @ class slides

    def test_class_problem(self):
        problem = KnapsackProblem(7, [1,2,3,4], [2,3,4,5])
        self.assertEqual(4, problem.size())
        self.assertEqual(7, problem.capacity)
        root = KnapsackSolution(problem)
        self.assertEqual(0, root.size())
        self.assertFalse(root.solution())
        self.assertEqual(0, root.eval())
        self.assertAlmostEqual( 9.00, root.estimation(), delta=0.01)
        self.assertAlmostEqual( 0.00, root.lower_bound(), delta=0.01)
        self.assertAlmostEqual(10.25, root.upper_bound(), delta=0.01)

        # Greedy
        greedy_solution, greedy_value = root.greedy()
        print(greedy_solution)
        self.assertAlmostEqual(10.25, greedy_value, delta=0.01)
        self.assertEqual(4, greedy_solution.size())
        self.assertTrue(greedy_solution.solution())
        self.assertEqual(6, greedy_solution.weight())
        self.assertEqual(9, greedy_solution.eval())
        self.assertEqual(9, greedy_solution.lower_bound())
        self.assertEqual(9, greedy_solution.upper_bound())
        self.assertEqual(1, greedy_solution.assignment[0])
        self.assertEqual(1, greedy_solution.assignment[1])
        self.assertEqual(1, greedy_solution.assignment[2])
        self.assertEqual(0, greedy_solution.assignment[3])

        # Branch&bound
        (benefit,solution) = search.branch_and_bound_max(root)
        self.assertEqual(10, benefit)
        self.assertEqual(4, solution.size())
        self.assertTrue(solution.solution())
        self.assertEqual(7, solution.weight())
        self.assertEqual(10, solution.eval())
        self.assertEqual(10, solution.lower_bound())
        self.assertEqual(10, solution.upper_bound())
        self.assertEqual(1, solution.assignment[0])
        self.assertEqual(1, solution.assignment[1])
        self.assertEqual(0, solution.assignment[2])
        self.assertEqual(1, solution.assignment[3])


if __name__ == '__main__':
    unittest.main()

 
'''
greedy_solution, greedy_value = root.greedy()
print(greedy_solution)

(c,s) = branch_and_bound_max(root)
print(f"Best cost = {c}")
print(f"Best solution = {s}")

(c,s) = branch_and_bound_max_no_greeedy(root)
print(f"Best cost = {c}")
print(f"Best solution = {s}")
'''