from __future__ import annotations
from queue import PriorityQueue
from math import inf


def branch_and_bound_min_no_greedy (root: Solution):
    lnv = PriorityQueue()
    lnv.put(root)  
    (c, s) = (inf, None)

    while not lnv.empty():
        x = lnv.get()
        print(f"- Best={c}, exploring {x}")
        if x.lower_bound() <= c:
            for y in x.children():
                if y.solution():
                    if y.eval()<c:
                        (c,s) = (y.eval(), y)
                else:
                    if y.lower_bound()<=c:
                        lnv.put(y)
    return (c,s)


def branch_and_bound_min (root: Solution):
    lnv = PriorityQueue()
    lnv.put(root)
    (c, s) = root.upper_bound_solution()              

    while not lnv.empty():
        x = lnv.get()
        print(f"- Best={c}, exploring {x}")
        if x.lower_bound() <= c:
            for y in x.children():
                if y.solution() and y.eval()<c:
                    (c,s) = (y.eval(), y)
                elif not y.solution() and y.lower_bound()<=c:
                    lnv.put(y)
                    (ctmp, stmp) = y.upper_bound_solution()
                    if ctmp<c:
                        (c,s) = (ctmp, stmp)
    return (c,s)


def branch_and_bound_max_no_greeedy(root: Solution):
    lnv = PriorityQueue()
    lnv.put(root)
    (c, s) = (-inf, None)

    while not lnv.empty():
        x = lnv.get()
        print(f"- Best={c}, exploring {x}")
        if x.upper_bound() >= c:
            for y in x.children():
                if y.solution():
                    if y.eval()>c:
                        (c,s) = (y.eval(), y)
                elif y.upper_bound()>=c:
                    lnv.put(y)
    return (c,s)


def branch_and_bound_max (root: Solution):
    lnv = PriorityQueue()
    lnv.put(root)
    (c, s) = root.lower_bound_solution()              

    while not lnv.empty():
        x = lnv.get()
        print(f"- Best={c}, exploring {x}")
        if x.upper_bound() >= c:
            for y in x.children():
                if y.solution() and y.eval()>c:
                    (c,s) = (y.eval(), y)
                elif not y.solution() and y.upper_bound()>=c:
                    lnv.put(y)
                    (ctmp, stmp) = y.lower_bound_solution()
                    if ctmp>c:
                        (c,s) = (ctmp, stmp)
    return (c,s)



# Base classes for representing problems and solutions

class Problem:
    '''
    Problem instance
    '''
    def root(self) -> Solution:
        '''
        Node value estimation
        '''
        pass


class Solution:
    '''
    Solution representation (i.e. node in graph)
    '''

    def __lt__(self, other): 
        '''
        Node comparison for the priority queue (minimization problem)
        '''
        return self.estimation()<other.estimation()

    def eval(self) -> float:
        '''
        Node evaluation
        '''
        pass

    def solution(self) -> bool:
        '''
        True if it represents a complete problem solution
        '''
        pass

    def children(self) -> list[Solution]:
        '''
        Node generation
        '''
        pass

    def estimation(self) -> float:
        '''
        Node value estimation
        '''

    def lower_bound(self) -> float:
        '''
        Lower bound
        '''
        pass

    def upper_bound(self) -> float:
        ''' 
        Upper bound
        '''
        pass

    def upper_bound_solution(self) -> (float, SolutionNode):
        ''' 
        Upper bound + solution, e.g. greedy algorithm for minimization problems
        '''
        pass

    def lower_bound_solution(self) -> (float, SolutionNode):
        ''' 
        Lower bound + solution, e.g. greedy algorithm for maximization problems
        '''
        pass
